/*****************************************************************************
 *
 *                        CalC (Calcium Calculator)
 *                             Victor Matveev
 *
 *                               fplot.cpp
 *
 *  Defines all plot objects, including mute (file dump) plots, xmgr/xmgrace
 *  command pipes and gl/Ygl X-windows graphs
 *
 ******************************************************************************/

#include "stdafx.h"  // for compatibility with Visual C++
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include <float.h>  // for compatibility with Visual C++
#include "Ygl.h"
#include <string.h>
#include <stdarg.h>
#include "vector.h"
#include "syntax.h"
#include "box.h"
#include "grid.h"
#include "field.h"
#include "table.h"
#include "peak.h"
#include "markov.h"
#include "interpol.h"
#include "gate.h"
#include "fplot.h"
#include "simulation.h"

#ifdef _YGL_INCLUDED_
void mouse_pause() {
    gflush(); while (getbutton(LEFTMOUSE)); 
}
#endif

int    XmgrPlot::XMGR_STEPS = 400;
int    PlotObj::UPDATE_STEPS = 600;
int    PlotObj::UPDATE_STEPS_1D = 200;
double PlotObj::UPDATE_ACCURACY = 0.001; 

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//        C L A S S   P L O T O B J E C T   I M P L E M E N T A T I O N
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@


void PlotObj::draw_grid(double x0, double x1, double y0, double y1, char lines)
{
#ifdef _YGL_INCLUDED_

  char xformat[20];
  char yformat[20];

  int xlines = int( double(plot_width) / double(pixelspergridline) + 0.5);
  int ylines = int( double(plot_height) / double(pixelspergridline) + 0.5);

  double dx = format_double(x0, x1, xlines, 1, xformat);
  double dy = format_double(y0, y1, ylines, 1, yformat);

  double yscale = double(plot_height) / double(y1 - y0);
  double xscale = double(plot_width) / double(x1 - x0);

  winset(winid);
  color(grid_color);
  sboxs(xmargin - 1, ymargin - 1, xmargin + plot_width + 1, ymargin + plot_height + 1);
  color(text_color);  

  double x, y;

  x = dx * floor(x0 / dx);  if (x0 > 0) x += dx;
  y = dy * floor(y0 / dy);  if (y0 > 0) y += dy;

  //fprintf(stderr,"x=%g x1-x0=%g dx=%g y=%g y0=%g y1=%g y1-y0=%g dy=%g \n", x, x1-x0, dx, y, y0, y1, y1-y0, dy);

  char s[50];
  int xx, yy;
  int count, limit = 100; // max number of grid divisions

  if (dx < 1.0e-9 ) dx = 1.0e-9;
  if (dy < 1.0e-9 ) dy = 1.0e-9;

  count = 0;
  while ( x <= x1 && count++ < limit)
    { 
    xx = int(xscale * (x - x0) + 0.5);
    if (lines) { 
               color(grid_color); 
               move2s(xx + xmargin, ymargin);
               draw2s(xx + xmargin, ymargin + plot_height); 
               }
    sprintf(s,xformat,x);
    cmov2s(xmargin + xx - 4, plot_height + ymargin + 4);
    color(text_color); charstr(s);
    x += dx;
    }

 count = 0;
 while ( y <= y1 && count++ < limit )
   { 
   yy = int(yscale * (y - y0) + 0.5); 
   if (lines) { 
              color(grid_color);
              move2s(xmargin, yy + ymargin);
              draw2s(xmargin + plot_width, ymargin + yy); 
              }
   sprintf(s,yformat,y);
   cmov2s(xmargin + plot_width + 3, ymargin + yy - 8);
   color(text_color); charstr(s);
   y += dy;
   }
#endif
}

//*******************************************************************************

void    StateDump::draw()
   {
   if ( Sim->Gates->Time < exportTime) { complete = false; return; }
   if ( complete ) return;
   complete = true;
   Sim->Export(fileName);
   };



//*******************************************************************************
//         C L A S S   P O I N T   P L O T   I M P L E M E N T A T I O N
//*******************************************************************************

PointPlot::PointPlot(double *ptr, double *tptr, char islog, double t, 
  const char *txt, const char *label, int hor, int ver) 
  : PlotObj(ptr, islog, txt, label)
{
#ifdef _YGL_INCLUDED_

  timeptr = tptr;

  count = 0;
 
  buffer = new double[UPDATE_STEPS + 2];  buffer[0] = get_value(0);
  times  = new double[UPDATE_STEPS + 2];  times[0] = *tptr;

  bufferSteps = UPDATE_STEPS;

  last = t + times[0];

  plot_width = hor;
  plot_height = ver;

  win_width = plot_width + 2 * xmargin + char_width * 8 + 3;
  win_height = plot_height + 2 * ymargin + char_height * 2 + 6;

  prefsize(win_width, win_height);
  winid = winopen(win_title);
  cmode();         
  gconfig(); qdevice(ESCKEY);
  putenv("YGL_PCM=1");

  set_colors();
  redraw();
#endif
}


//*************************************************************************************

void PointPlot::set_colors()
{
#ifdef _YGL_INCLUDED_
  mapcolor(grid_color,  255,255,249);
  mapcolor(bgr_color,   236,230,236);
  mapcolor(plot_color,  210,205,210);
  mapcolor(text_color,  0, 0, 0);
  mapcolor(label_color, 0, 0, 255);
  mapcolor(line_color,  255, 0, 255);
#endif
}

//*************************************************************************************
//                         T I M E   P L O T   D R A W
//*************************************************************************************

void PointPlot::draw()
{
#ifdef _YGL_INCLUDED_

int    x, xn, y, yn;
char   txt[200], xform[80], xtxt[80], format[80];
char   mmform[30], mintxt[30], maxtxt[40], currtxt[40];

static double tscale = double(plot_width) / last;
double value = get_value(0);

  double t = *timeptr;

  if (t <= times[0]) { fmin = fmax = get_value(0);  fmax += 1.0e-12; }

  //fprintf(stdout,"%d, %lg %lg\n",count,t,get_value(0));

  if (t < times[count]) {
    fmin = fmax = get_value(0);  fmax += 1.0e-12;
    for (count = 0; t > times[count]; count++) 
      if (buffer[count] < fmin) fmin = buffer[count];
      else if (buffer[count] > fmax) fmax = buffer[count];
    if (count) count--;
  }

  if  ( count && (int( t * tscale ) == int( times[count] * tscale ) ) ) {
    if ( value < fmin )  { fmin = buffer[count] = value; times[count] = t; } // overwrite old values with
    if ( value > fmax )  { fmax = buffer[count] = value; times[count] = t; }  // extreme values
    return;
  }

  if ( value < fmin )  fmin = value;
  if ( value > fmax )  fmax = value;   

  if (count < bufferSteps && t > times[count])  ++count;
  else if (count >= bufferSteps) {                           // if buffer overflow
    double *newtimes = new double[bufferSteps + 400 + 2];
    double *newbuffer = new double[bufferSteps + 400 + 2];
    bufferSteps += 400;
    for (int i = 0; i <= bufferSteps; i++) { newtimes[i] = times[i]; newbuffer[i] = buffer[i]; }
    delete [] times; delete [] buffer;
    times = newtimes; buffer = newbuffer;
    count ++;
  } 

  times[count] = t;
  buffer[count] = value;

  winset(winid);
  color(bgr_color);  clear();
  color(plot_color);
  sboxfs(xmargin, ymargin, xmargin + plot_width, ymargin + plot_height);

  //printf("%g %g %e %e\n",times[0], last, fmin, fmax);
  draw_grid(times[0], last, fmin, fmax, 1);

  int ychar = ymargin + plot_height + getheight() + 7; 
  int xchar = xmargin - 4;

  color(bgr_color);
  sboxfs(xchar - 1, ychar - 1, win_width, win_height);

  strcpy(xform, x_label);
  format_double(times[count-1], times[count], 1, 1, format);
  strcat(xform,format);

  format_double(fmin, fmax, 1, 3, mmform);
  if (fmin == 0.0) strcpy(mintxt,"0"); else sprintf(mintxt, mmform, fmin);
  if (fmax == 0.0) strcpy(maxtxt,"0"); else sprintf(maxtxt, mmform, fmax);
  sprintf(xtxt, xform, times[count]);
  format_double(buffer[count-1], value, 1, 3, format);
  sprintf(currtxt,format,value);

  sprintf(txt, "%s min=%s max=%s current=%s", xtxt, mintxt, maxtxt, currtxt);
  cmov2s(xchar, ychar);  color(label_color);
  charstr(txt);

  color(line_color);
  linewidth(2);
  double xscale = double(plot_width) / (last - times[0]);
  double yscale = double(plot_height) / (fmax - fmin);

  pnt2s(xmargin, ymargin + int( yscale * (buffer[0] - fmin) + 0.5 ) );
  yn = x = 0;

  for (int j = 1; j <= count; j++)
    {
    xn = xmargin + int( xscale * (times[j] - times[0]) + 0.5 );
    y  = ymargin + int( yscale * (buffer[j] - fmin) + 0.5 );
    if (y > yn) yn = y;
    if (xn > x) { draw2s(xn, yn); yn = 0; }
    x = xn;
    }

  linewidth(1);
  mouse_pause();
#endif
}

//*************************************************************************************
//                M U T E    P L O T   D R A W
//*************************************************************************************

MutePointPlot::MutePointPlot(double *ptr, double *tptr, char islog,  double t, 
               const char *fname, const char *txt) : PlotObj(ptr, islog, txt)
    {
    tscale = double(UPDATE_STEPS) / t;
    fmin = fold = fmax = get_value();
    x_temp = x_value = -1e12;
    time = tptr;
    f_temp = 0.0;
    strcpy(fileName,fname);
    file = fopenAssure(fname, "w", "plot", win_title);
    }

//*************************************************************************************

void    MutePointPlot::draw()
   {
   bool redrawFlag, newValue;

   double f = get_value();
   if (f > fmax) fmax = f; 
   else if (f < fmin) fmin = f;
   if ( *time < x_value && file != stdout  && file != stderr ) {  // do a back step
     double x = 0, y = 0;
     fclose(file);
     FILE *temp = fopenAssure(fileName, "r", "plot", win_title);
     file = fopenAssure("plotTempFile", "w", "plot",  win_title);
     while (!feof(temp)) {
       fscanf(temp,"%lf %lf",&x, &y);
       if (feof(temp) || x >= *time) break;
       x_value = x; fold = y;
       fprintf(file,"%.10g %.10g\n",x, y);
     }
     fclose(temp); fclose(file);
     temp = fopenAssure("plotTempFile", "r", "plot", win_title);
     file = fopenAssure(fileName, "w", "plot",  win_title);
     while (!feof(temp)) {
       fscanf(temp,"%lf %lf",&x, &y);
       if (feof(temp)) break;
       fprintf(file,"%.10g %.10g\n",x, y);
     }
     fclose(temp);
   }
   else {
       redrawFlag = ( int( *time * tscale ) == int( x_value * tscale) ) ? false : true;
       newValue = ( fabs(f - fold) < (fmax - fmin) * UPDATE_ACCURACY ) ? false : true;
       if ( !redrawFlag && !newValue ) { x_temp = *time; f_temp = f; return; }
       if (newValue)
	 if (x_temp > x_value)  // if we didn't draw this point before
           if ( fabs(f_temp - f) > (fmax - fmin) * UPDATE_ACCURACY ) // and we deviated from it
              fprintf(file, "%.10g %.10g \n", x_temp, f_temp);
     }

   fold = f_temp = f;
   x_value = x_temp = *time; 
   fprintf(file, "%.10g %.10g \n", *time, get_value() );
   fflush(file);
   }

//*************************************************************************************

void MutePlot1D::draw() {

   FILE *file;
   double t = field->Time;
   long ind = field_index;
   int  i;

   if (UPDATE_STEPS_1D == 1) {
     //fprintf(stderr,"t=%g export=%g t-e=%e fname=%s \n",t, exportTime, t - exportTime, fileName);
       if (t + 1e-20 < exportTime) { complete = false; return; }
       if ( complete ) return;
       complete = true;
       file = fopenAssure(fileName, "w", "plot", win_title);
       for (i = 0; i < num; i++, ind+=incr)
            fprintf(file,"%.6g %.9g \n", coord[i], get_value(ind));
       fclose(file);
   }
   else if ( t < x_value && !equal(fileName,"stdout")  && !equal(fileName,"stderr") ) {  // do a back step
     int  BUFF_SIZE = 256;
     char buffer[256];
     double x, y, z;
     FILE *temp = fopenAssure("plotTempFile", "w", "plot",  win_title);
     file = fopenAssure(fileName, "r", "plot",  win_title);
     while (!feof(file)) {
       fgets(buffer, BUFF_SIZE, file);
       if (strlen(buffer) > 4) {
         sscanf(buffer,"%lf %lf %lf",&x, &y, &z);
         if (feof(file) || x >= t) break;
         x_value = x;
       }
       fprintf(temp,"%s", buffer);
     }
     fclose(temp); fclose(file);
     temp = fopenAssure(fileName, "w", "plot",  win_title);
     file = fopenAssure("plotTempFile", "r", "plot",  win_title);
     while (!feof(file)) {
       fgets(buffer, BUFF_SIZE, file);
       if (feof(file) ) break;
       fprintf(temp,"%s", buffer);
     }
     fclose(temp); fclose(file);
   }
   else  if ( int(t * tscale) > int(x_value * tscale) ) {
     x_value = t;
     file = fopenAssure(fileName, "a", "plot", win_title);
     for (int i = 0; i < num; i++, ind += incr) 
        fprintf(file,"%.10g %.10g %.10g\n", t, coord[i], get_value(ind));
     fprintf(file,"\n");
     fclose(file);
   }

}

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//         C L A S S   2 D - F I E L D   P L O T   I M P L E M E N T A T I O N
//*******************************************************************************

FieldPlot2D::FieldPlot2D(FieldObj *f, char islog, const char *dir, double coord, const char *label) : 
FieldPlotObj(f, islog, "", label)
{
 

  if ( equal(dir,LABEL_DIM1) && DIMENSIONALITY == 3) {                 // y-z plane
            grid1 = FieldObj::ygrid;  grid2  = FieldObj::zgrid;
            coord1= FieldObj::ycoord; coord2 = FieldObj::zcoord;
            incr1 = FieldObj::xsize;  incr2  = FieldObj::xysize; 
            num1  = FieldObj::ysize;  num2   = FieldObj::zsize;
            coord = FieldObj::Grid->nearest_location(FieldObj::xcoord, FieldObj::xgrid, FieldObj::xsize, coord);
            field_index = FieldObj::Grid->location_to_index(coord, coord1[0], coord2[0]);
	    sprintf(win_title,"%s[%.3g,%s,%s]", f->ID, coord, LABEL_DIM2, LABEL_DIM3);
  } else 
    if ( equal(dir,LABEL_DIM2) && DIMENSIONALITY == 3 ) {             // x-z plane
            grid1  = FieldObj::xgrid;   grid2  = FieldObj::zgrid;
            coord1 = FieldObj::xcoord;  coord2 = FieldObj::zcoord;
            incr1  = 1;                 incr2  = FieldObj::xysize; 
            num1   = FieldObj::xsize;   num2   = FieldObj::zsize;
            coord  = FieldObj::Grid->nearest_location(FieldObj::ycoord, FieldObj::ygrid, FieldObj::ysize, coord);
            field_index = FieldObj::Grid->location_to_index(coord1[0], coord, coord2[0]);
            sprintf(win_title,"%s[%s,%.3g,%s]", f->ID, LABEL_DIM1, coord, LABEL_DIM3);
    } else                                                            // x-y plane
  if  ( (equal(dir,LABEL_DIM3) && DIMENSIONALITY > 1) || DIMENSIONALITY == 2 ) {
            grid1  = FieldObj::xgrid;  grid2  = FieldObj::ygrid;
            coord1 = FieldObj::xcoord; coord2 = FieldObj::ycoord;
            incr1  = 1;                incr2  = FieldObj::xsize; 
            num1   = FieldObj::xsize;  num2   = FieldObj::ysize;
            coord  = FieldObj::Grid->nearest_location(FieldObj::zcoord, FieldObj::zgrid, FieldObj::zsize, coord);
            field_index = FieldObj::Grid->location_to_index(coord1[0], coord2[0], coord);
	    if (DIMENSIONALITY == 2) sprintf(win_title,"%s[%s,%s]", f->ID, LABEL_DIM1, LABEL_DIM2);
	      else sprintf(win_title,"%s[%s,%s,%.3g]", f->ID, LABEL_DIM1, LABEL_DIM2, coord);
  }
  else throw makeMessage("Cannot interpret direction \"%s\" in 2D plot",dir);
}


//*************************************************************************************

void FieldPlot2D::get_range()
{
long   ind;
double f;

  fmin = get_value(field_index);
  fmax = get_value(field_index);


  for (int j = 0; j < num2; j++)
    {
    for (int i = 0; i < num1; i++)  
      {
      ind = field_index + j * incr2 + i * incr1;
      if (field->ptype[ind] & REGION_MASK)
        {
	f = get_value(ind);
        if (f < fmin) fmin = f; else if (f > fmax) fmax = f;
	}
      }
    }
  if ( (fabs( (fmax-fmin)/(fmax+fmin+1.0e-12)) < 1e-16 ) || (fmax <= fmin) )
     fmax = (fmin + 1e-8);
}
//*******************************************************************************
//            C L A S S   2 D - P L O T   I M P L E M E N T A T I O N
//*******************************************************************************

Plot2D::Plot2D(FieldObj *f, char islog, const char *dir, double coord, double total, int hor, int ver, 
               double min, double max, const char *label) : FieldPlot2D(f, islog, dir, coord, label) 
{
#ifdef _YGL_INCLUDED_

tscale = double(UPDATE_STEPS) / total;

if (max > min)
  {
  min_fixed = min; max_fixed = max;
  fixed = 1;
  df_dcol_fixed = (max_fixed - min_fixed) / double(color_num);
  }
else fixed = 0;

    double f1 = double(hor) / double(num1+1);
    double f2 = double(ver) / double(num2+1);
    double factor = ( (f1 > f2) ? f2 : f1);

    plot_width  = int(factor * (num1 + 1) + 0.5);
    plot_height = int(factor * (num2 + 1) + 0.5);

    //    plot_width  = hor;
    //   plot_height = ver;

    win_width  = plot_width  + 2 * xmargin + char_width * 8 + 4;
    win_height = plot_height + 2 * ymargin + dy_color_panel + 3 * char_height + 10; 
    
    prefsize(win_width, win_height);
    winid = winopen(win_title);
    cmode();         
    gconfig();  qdevice(ESCKEY);
    putenv("YGL_PCM=1");
    set_colors();
    redraw();
#endif
}

//*************************************************************************************

void Plot2D::redraw()
{
#ifdef _YGL_INCLUDED_
  winset(winid);
  color(bgr_color);
  clear();
  color(plot_color);
  sboxfs(xmargin, ymargin, xmargin + plot_width, ymargin + plot_height);
  draw_grid(coord1[0], coord1[num1-1], coord2[0], coord2[num2-1], 0);
  x_value = 0; // field->Time - 0.000001;
  draw();
#endif
}

//*************************************************************************************

void Plot2D::set_colors()
{
#ifdef _YGL_INCLUDED_
  mapcolor(grid_color,  200, 150, 150);
  mapcolor(bgr_color,   236, 226, 236);
  mapcolor(plot_color,  210, 230, 220);
  mapcolor(text_color,  0,   0,   0);
  mapcolor(label_color, 255, 0,   0);

  int red, green, blue;
  color1 = 255 - color_num;

  for(int i=color1; i<= color1 + color_num; i++)  
      {
      red = green = int( 255.0 * (i - color1) / double(color_num) );
      blue = 255 - red;
      mapcolor(i, red, green, blue);
      }
#endif
}

//*************************************************************************************

void Plot2D::color_panel()
{
#ifdef _YGL_INCLUDED_

int  y0, y1, x0, x1;
char mintxt[40], smin[40];
char maxtxt[40], smax[40];
char form[40];
 
  y0 = ymargin + plot_height + getheight() + 10;
  y1 = y0 + dy_color_panel;
  x0 = xmargin - 4;

  color(bgr_color);
  sboxfs(x0 - 1, y0 - 1, win_width, y1 + 1);

  get_range();

  format_double(fmin, fmax, 1, 3, form);  
  if (fmin == 0.0) strcpy(smin,"0"); else sprintf(smin, form, fmin);
  if (fmax == 0.0) strcpy(smax,"0"); else sprintf(smax, form, fmax);

  sprintf(mintxt, "min=%s", smin);
  sprintf(maxtxt, "max=%s", smax);

  color(text_color);
  cmov2(x0, y0), charstr(mintxt);
 
  x0 = x0 + strwidth(mintxt) + 4;  // x0 coordinate for color panel
  x1 = win_width - (xmargin - 4) - strwidth(maxtxt);  // coordinate for "fmax="

  cmov2(x1, y0), charstr(maxtxt);

  x1 -= 6;  // x1 coordinate for color panel

  double dfdx = (fmax - fmin) / double(x1 - x0);
  double f;

  color(grid_color);
  sboxfs(x0-1, y0-1, x1 + 1, y1 + 1);

  for (int xn = x0; xn <= x1; xn++ )
    {
    f = fmin + dfdx * (xn - x0);
    color(get_color(f)); 
    pnt2s(xn, y0); draw2s(xn,y1);
    }
#endif
}

//*************************************************************************************
//                          P L O T  2 D   D R A W
//*************************************************************************************

void Plot2D::draw()
{
#ifdef _YGL_INCLUDED_

long   ind;
char   text1[80], text2[80];
static char form[80];

  double t = field->Time;
  //  fprintf(stdlog, "t=%g x_value=%g tscale=%g \n",t,x_value,tscale);
  if ( int(t * tscale) == int(x_value * tscale) ) return; 

  winset(winid);
  color_panel();

  int ychar = ymargin + plot_height + getheight() + dy_color_panel + 15; 
  int xchar = xmargin - 4;

  color(bgr_color);
  sboxfs(xchar - 1, ychar - 1, win_width, win_height);

  format_double(x_value, t, 1, 2, form);
  x_value = t;
  sprintf(text1, form, x_value);
  sprintf(text2, "%s%s", x_label, text1);
  cmov2s(xchar, ychar);  color(label_color);
  charstr(text2);

  double scale1 = double(plot_width) / (coord1[num1-1] - coord1[0] + 0.5 * ( grid1[0] + grid1[num1-1]) );
  double scale2 = double(plot_height) / (coord2[num2-1] - coord2[0] + 0.5 * ( grid2[0] + grid2[num2-1]) );

  int c;
  int    x, y, xx, yy;
  double dx, dy;
  y = ymargin;

  for (int j = 0; j < num2; j++)
    {
    x = xmargin;
    dy = coord2[j] - coord2[0] + 0.5 * ( grid2[0] + grid2[ j ? j - 1 : 0] ); 
    yy = int( dy * scale2 + 0.5) + ymargin;
    for (int i = 0; i < num1; i++)  
      {
      ind = field_index + j * incr2 + i * incr1;
      dx = coord1[i] - coord1[0] + 0.5 * ( grid1[0] + grid1[ i ? i - 1 : 0] ); 
      xx = int(dx * scale1 + 0.5) + xmargin;
      if (field->ptype[ind] & REGION_MASK)
        {
        c = get_color( get_value(ind) );
        if (c > 0) { color( c ); sboxfs(x, y, xx, yy); }
        }
      x = xx;
      }
    y = yy;
    } 
  mouse_pause();
#endif
}

//*************************************************************************************

int Plot2D::get_color(double f)
{
int c;
if (! fixed)
  {
  c = color1 + int( (f - fmin) / df_dcol + 0.5 );
  if (c < color1 || c > 255) 
    throw makeMessage("Plot2D::get_color: color out of range: f=%g fmin=%g fmax=%g\n", f, fmin, fmax);
  }
else
  {
  c = color1 + int( (f - min_fixed) / df_dcol_fixed + 0.5 );
  if (c > 255) c = color_num; 
    else if (c < color1) c = color1; 
  }
return c;
}

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//*******************************************************************************
//       B A S E   F I E L D   P L O T   1 D   I M P L E M E N T A T I O N
//*******************************************************************************

double FieldPlotObj::get_value(long ind)
   {
   double v = fptr[ind];
   long   pointType = field->ptype[ind];
   if ( !(pointType & VARY_MASK ) ) {
     //     fprintf(stderr,"fmin=%g \n",fmin);
     v = (pointType & DIRICHLET) ? field->bc_coef[pointType >> XSHIFT] : fmin; }
   if (!log_plot) return v;
      else return ( (v <= 0.0) ? 0.0 : log(v)/log(10.0) );
   }

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

FieldPlot1D::FieldPlot1D(FieldObj *f, char islog, const char *dir, double coord1, double coord2,
               const char *label) :  FieldPlotObj(f, islog, "", label)
{
  if ( equal(dir, LABEL_DIM1) ) {
            grid = FieldObj::xgrid; coord = FieldObj::xcoord; 
            incr = 1;  num = FieldObj::xsize;
            coord1 = FieldObj::Grid->nearest_location(FieldObj::ycoord, FieldObj::ygrid, FieldObj::ysize, coord1);
            coord2 = FieldObj::Grid->nearest_location(FieldObj::zcoord, FieldObj::zgrid, FieldObj::zsize, coord2);
            field_index = FieldObj::Grid->location_to_index(coord[0], coord1, coord2);
	    if (DIMENSIONALITY == 1) sprintf(win_title,"%s[%s]", f->ID, LABEL_DIM1);
	    else if (DIMENSIONALITY == 2) sprintf(win_title,"%s[%s,%.3g]", f->ID, LABEL_DIM1, coord1);
	    else sprintf(win_title,"%s[%s,%.3g,%.3g]", f->ID, LABEL_DIM1, coord1, coord2);
 } else
   if ( equal(dir, LABEL_DIM2) ) {
            grid = FieldObj::ygrid; coord = FieldObj::ycoord;
            incr = FieldObj::xsize; num = FieldObj::ysize;
            coord1 = FieldObj::Grid->nearest_location(FieldObj::xcoord, FieldObj::xgrid, FieldObj::xsize, coord1);
            coord2 = FieldObj::Grid->nearest_location(FieldObj::zcoord, FieldObj::zgrid, FieldObj::zsize, coord2);
            field_index = FieldObj::Grid->location_to_index(coord1, coord[0], coord2);
	    if (DIMENSIONALITY == 2) sprintf(win_title,"%s[%.3g,%s]", f->ID, coord1, LABEL_DIM2);
	    else sprintf(win_title,"%s[%.3g,%s,%.3g]", f->ID, coord1, LABEL_DIM2, coord2);
 } else 
  if  ( equal(dir, LABEL_DIM3) ) {
            grid = FieldObj::zgrid;  coord = FieldObj::zcoord;
            incr = FieldObj::xysize; num = FieldObj::zsize;
            coord1 = FieldObj::Grid->nearest_location(FieldObj::xcoord, FieldObj::xgrid, FieldObj::xsize, coord1);
            coord2 = FieldObj::Grid->nearest_location(FieldObj::ycoord, FieldObj::ygrid, FieldObj::ysize, coord2);
            field_index = FieldObj::Grid->location_to_index(coord1, coord2, coord[0]);
	    sprintf(win_title,"%s[%.3g,%.3g,%s]", f->ID, coord1, coord2, LABEL_DIM3);
  }
  else throw makeMessage("Cannot interpret direction \"%s\" in 1D plot", dir);
}

//*************************************************************************************

void FieldPlot1D::get_range()
{
long   ind;
double f;
bool   flag = true;

  for (int j = 0; j < num; j++)  {
    ind = field_index + j * incr;
    if (field->ptype[ind] & REGION_MASK)  
      if (flag) {
        fmin = get_value(field_index);
        fmax = get_value(field_index); 
        flag = false;
      }
      else {
        f = get_value(ind);
        if (f < fmin) fmin = f;  
          else if (f > fmax) fmax = f;
      }
  }
    
  //fprintf(stderr,"fmin=%g fmax=%g \n",fmin,fmax);

  if ( (fabs((fmax - fmin) / (fmax+fmin+1.0e-12)) < 1.0e-16) || (fmax <= fmin) )
     fmax = fmin + 1e-8;
}

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//*******************************************************************************
//                   P L O T   1 D   I M P L E M E N T A T I O N
//*******************************************************************************

Plot1D::Plot1D(FieldObj *f, char islog, const char *dir, double coord1, double coord2,
               double total, int hor, int ver, const char *label) :  
        FieldPlot1D(f, islog, dir, coord1, coord2, label)
{
#ifdef _YGL_INCLUDED_
  tscale = double(UPDATE_STEPS) / total;

  plot_width = hor;
  plot_height = ver;

  win_width = plot_width + 2 * xmargin + char_width * 8 + 3;
  win_height = plot_height + 2 * ymargin + char_height * 2 + 6;
  prefsize(win_width, win_height);
  winid = winopen(win_title);
  cmode();  gconfig(); qdevice(ESCKEY);
  putenv("YGL_PCM=1");
  set_colors();
  redraw();
#endif
}


//*************************************************************************************

void Plot1D::set_colors()
{
#ifdef _YGL_INCLUDED_
  mapcolor(grid_color,  255,255,249);
  mapcolor(bgr_color,   236,230,236);
  mapcolor(plot_color,  210,205,210);
  mapcolor(text_color,  0, 0, 0);
  mapcolor(label_color, 0, 0, 255);
  mapcolor(line_color,  255, 0, 255);
#endif
}

//*************************************************************************************
//                           P L O T   1 D   D R A W
//*************************************************************************************

void Plot1D::draw()
{
#ifdef _YGL_INCLUDED_

long   ind;
int    x, y;
char   text[80], xtxt[80], form[80];
char   smin[40], smax[40], mintxt[40], maxtxt[40];

  double t = field->Time;
  if ( int(tscale * x_value) == int(tscale * t) ) return;

  int iLeft = 0, iRight = num-1;
  while ( !(field->ptype[field_index+iLeft*incr] & REGION_MASK) ) iLeft++;
  while ( !(field->ptype[field_index+iRight*incr] & REGION_MASK) ) iRight--;
  
  winset(winid);
  color(bgr_color);  clear();
  color(plot_color);
  sboxfs(xmargin, ymargin, xmargin + plot_width, ymargin + plot_height);
  get_range();
  draw_grid(coord[iLeft], coord[iRight], fmin, fmax, 1);

  int ychar = ymargin + plot_height + getheight() + 7; 
  int xchar = xmargin - 4;

  color(bgr_color);
  sboxfs(xchar - 1, ychar - 1, win_width, win_height);

  format_double(fmin, fmax, 1, 3, form);  
  if (fmin == 0.0) strcpy(smin,"0"); else sprintf(smin, form, fmin);
  if (fmax == 0.0) strcpy(smax,"0"); else sprintf(smax, form, fmax);
  sprintf(mintxt, "min=%s", smin);
  sprintf(maxtxt, "max=%s", smax);

  format_double(x_value, t, 1, 2, form);
  x_value = t;
  sprintf(xtxt, form, x_value);
  sprintf(text, "%s%s %s %s", x_label, xtxt, mintxt, maxtxt);
  cmov2s(xchar, ychar);  color(label_color);
  charstr(text);

  color(line_color);
  linewidth(2);
  x = xmargin;
  ind = field_index + iLeft * incr;
  pnt2s(xmargin, 
        ymargin + int( plot_height * (get_value(ind) - fmin) / (fmax - fmin) + 0.5 ));

  double scale = double(plot_width) / (coord[iRight] - coord[iLeft]);

  for (int j = iLeft; j <= iRight; j++, ind += incr)
    {
    x = xmargin + int( (coord[j] - coord[iLeft]) * scale + 0.5);
    y = ymargin;
    y += int( plot_height * (get_value(ind) - fmin) / (fmax - fmin) + 0.5 );
    draw2s(x,y);
    }

  linewidth(1);
  mouse_pause();
#endif
}

//*******************************************************************************
//        C L A S S   R A S T E R    P L O T   I M P L E M E N T A T I O N
//*******************************************************************************

RasterPlot::RasterPlot(FieldObj *f, char islog, const char *dir, double coord1, double coord2, double t,
           int hor, int ver,  double min, double max, 
           const char *label) : FieldPlot1D(f, islog, dir, coord1, coord2, label)
{
#ifdef _YGL_INCLUDED_
if (max > min)
  {
  min_fixed = min; max_fixed = max;
  fixed = 1;
  df_dcol_fixed = (max_fixed - min_fixed) / double(color_num);
  }
else fixed = 0;

  fmin = get_value(field_index);
  fmax = get_value(field_index); 

  count = 0; 
  raster = new double[(UPDATE_STEPS + 2) * num + 1];
  times  = new double[UPDATE_STEPS + 2];
  
  for (int j = 0; j < num; j++) raster[j] = get_value(field_index + j * incr);
  times[0] = f -> Time; 
  last = times[0] + t;
 
  plot_width = hor; plot_height = ver;

  win_width = plot_width + 2 * xmargin + char_width * 8 + 4;
  win_height = plot_height + 2 * ymargin + dy_color_panel + 3 * char_height + 10;

  prefsize(win_width, win_height);
  winid = winopen(win_title);
  cmode(); gconfig(); qdevice(ESCKEY);
  putenv("YGL_PCM=1");
  set_colors();
  redraw();
#endif
}

//*************************************************************************************

void RasterPlot::redraw()
{
#ifdef _YGL_INCLUDED_
  winset(winid);
  color(bgr_color);
  clear();
  color(plot_color);
  sboxfs(xmargin, ymargin, xmargin + plot_width, ymargin + plot_height);
  draw_grid(times[0], last, coord[0], coord[num-1], 0);
  if (count) count --;
  draw();
#endif
}

//*************************************************************************************

void RasterPlot::set_colors()
{
#ifdef _YGL_INCLUDED_
  mapcolor(grid_color,  200, 150, 150);
  mapcolor(bgr_color,   236,226,236);
  mapcolor(plot_color,  210, 205, 210);
  mapcolor(text_color,  0, 0, 0);
  mapcolor(label_color, 255, 0, 0);

  int red, green, blue;
  color1 = 255 - color_num;

  for(int i=color1; i<= color1 + color_num; i++)  
      {
      red = green = int( 255.0 * (i - color1) / double(color_num) );
      blue = 255 - red;
      mapcolor(i, red, green, blue);
      }
#endif
}

//*************************************************************************************

void RasterPlot::get_range()
{
  FieldPlot1D::get_range();
  df_dcol = (fmax - fmin) / double(color_num);
}

//*************************************************************************************

void RasterPlot::color_panel()
{
#ifdef _YGL_INCLUDED_

int  y0, y1, x0, x1;
char mintxt[40], smin[40];
char maxtxt[40], smax[40];
char form[40];
 
  y0 = ymargin + plot_height + getheight() + 10;
  y1 = y0 + dy_color_panel;
  x0 = xmargin - 4;

  color(bgr_color);
  sboxfs(x0 - 1, y0 - 1, win_width, y1 + 1);

  format_double(fmin, fmax, 1, 3, form);  
  if (fmin == 0.0) strcpy(smin,"0"); else sprintf(smin, form, fmin);
  if (fmax == 0.0) strcpy(smax,"0"); else sprintf(smax, form, fmax);
  sprintf(mintxt, "min=%s", smin);
  sprintf(maxtxt, "max=%s", smax);

  color(text_color);
  cmov2(x0, y0), charstr(mintxt);
 
  x0 = x0 + strwidth(mintxt) + 4;  // x0 coordinate for color panel
  x1 = win_width - (xmargin - 4) - strwidth(maxtxt);  // coordinate for "fmax="

  cmov2(x1, y0), charstr(maxtxt);

  x1 -= 6;  // x1 coordinate for color panel

  double dfdx = (fmax - fmin) / double(x1 - x0);
  double f;

  color(grid_color);
  sboxfs(x0-1, y0-1, x1 + 1, y1 + 1);

  for (int xn = x0; xn <= x1; xn++ )
    {
    f = fmin + dfdx * (xn - x0);
    color(get_color(f)); 
    pnt2s(xn, y0); draw2s(xn,y1);
    }
#endif
}

//*************************************************************************************
//                         R A S T E R   P L O T   D R A W
//*************************************************************************************

void RasterPlot::draw()
{
#ifdef _YGL_INCLUDED_

char   text1[80], text2[80], form[80];
int    i, j;

  double t = field->Time;

  if (count > 0) 
    {
    while (t < times[count] && count) count--;
    if ( int( times[count] * double(plot_width) / last )
       ==  int( t * double(plot_width) / last ) ) return;
    }

  if (t > times[count] && count < UPDATE_STEPS )  times[++count] = t;
  else if (count >= UPDATE_STEPS) 
    throw makeMessage("RasterPlot::draw: buffer overflow [steps=%d]\n",UPDATE_STEPS);

  int iLeft = 0, iRight = num-1;
  while ( !(field->ptype[field_index+iLeft*incr] & REGION_MASK) ) iLeft++;
  while ( !(field->ptype[field_index+iRight*incr] & REGION_MASK) ) iRight--;

  for (i = iLeft; i <= iRight; i++)
      raster[ count * num + i ] = get_value(field_index + i * incr);
  get_range();  
  winset(winid);
  color_panel();

  int ychar = ymargin + plot_height + getheight() + dy_color_panel + 15; 
  int xchar = xmargin - 4;

  color(bgr_color);
  sboxfs(xchar - 1, ychar - 1, win_width, win_height);

  format_double(times[count-1], times[count], 1, 1, form);
  sprintf(text1, form, times[count]);
  sprintf(text2, "%s%s", x_label, text1);
  cmov2s(xchar, ychar);  color(label_color);
  charstr(text2);

  double scale1 = double(plot_width) / (last - times[0]);
  double scale2 = double(plot_height) / (coord[iRight] - coord[iLeft]);

  int    x, y, xx, yy;
  double dy;

  y = ymargin;

  if (count == 0) return;
 
  for (j = iLeft; j <= iRight; j++)
    {
    x = xmargin;
    dy = ( j < iRight ? coord[j+1] - 0.5*grid[j] : coord[iRight] ) - coord[iLeft]; 
    yy = int( dy * scale2 + 0.5) + ymargin;
    for (i = 0; i < count; i++)
      {  
      xx = xmargin + int( (0.5 * (times[i] + times[i+1]) - times[0]) * scale1 + 0.5);
      if (xx > x) if (field->ptype[field_index + j * incr] & REGION_MASK)
        { 
        color( get_color( raster[i * num + j] ) );
        sboxfs(x, y, xx, yy);
        }
      x = xx;
      }
    y = yy;
    }
  mouse_pause();
#endif
}

//*************************************************************************************

int RasterPlot::get_color(double f)
{
int c;
if (! fixed)
  {
  c = color1 + int( (f - fmin) / df_dcol + 0.5 );
  if (c < color1 || c > 255) {  }
  if (c > 255) c = color_num; 
    else if (c < color1) c = color1; 
  }
else
  {
  c = color1 + int( (f - min_fixed) / df_dcol_fixed + 0.5 );
  if (c > 255) c = color_num; 
    else if (c < color1) c = color1; 
  }
return c;
}

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//*******************************************************************************
//             B A S E   C L A S S   X M G R   P I P E   P L O T
//*******************************************************************************

 int XmgrPlot::graph_num = 0;
 int XmgrPlot::graph_count = 0;
 int XmgrPlot::rows = 0;
 int XmgrPlot::cols = 0;
 FILE *XmgrPlot::file = stdout;

//*******************************************************************************

  void XmgrPlot::init(TokenString &TS, int gnum, FILE *f, int rs, int cs)
    {
    if ( !(graph_num = gnum) ) return;
    file = f;
    rows = rs; cols = cs;
    if (rows * cols < graph_num) {
      rows = int( sqrt(double(graph_num)) + 0.5 );
      cols = int( ceil( double(graph_num) / double(rows) ) );
    }
    double dx = 1.0 / double(cols);
    double dy = 1.0 / double(rows);  

    TS.get_int_param("plot.update.steps", &XMGR_STEPS);

    //    fprintf(file, "@ version 40102 \n");
    //    fprintf(file, "@ page layout free \n");      
    fprintf(file, "@ TIMESTAMP on \n");
    fprintf(file, "@ TIMESTAMP CHAR SIZE 0.6\n");

    for (int i = 0; i < rows; i++)
      {
      double y0 = 1.0 - (i + 1) * dy + dymargin / rows;
      double y1 = 1.0 - i * dy - dymargin / rows;
      for (int j = 0; j < cols; j++)
	{ 
        double x0 = j * dx + 1.5 * dxmargin / cols;
        double x1 = (j + 1) * dx - 0.5 * dxmargin / cols;
	int  graph = i * cols + j;
        if (graph >= graph_num) break;
        fprintf(file, "@with g%d\n", graph);
        fprintf(file, "@g%d on\n", graph);
        fprintf(file, "@ world ymin 0.000001\n");
        fprintf(file, "@ world ymax 1.000000\n");
        fprintf(file, "@ view xmin %g\n", x0);
        fprintf(file, "@ view xmax %g\n", x1);
        fprintf(file, "@ view ymin %g\n", y0);
        fprintf(file, "@ view ymax %g\n", y1);
	}
      }

    processStrings(TS);
  }
		

//*******************************************************************************

 XmgrPlot::XmgrPlot(int sets, char is_log, const char *xlabel)
  {
  graph_id = graph_count;

  if (is_log) logScale = true; 
  else logScale = false;
 
  if ( ++graph_count > graph_num )  
     throw makeMessage("XmgrPlot::XmgrPlot: too many graphs: [%d]>[%d]\n", graph_count, graph_num);

  set_num = sets;
  set_count = 0;

  fprintf(file, "@with g%d\n", graph_id);
  fprintf(file, "@g%d on\n", graph_id);
  if (is_log) fprintf(file, "@g%d type logy\n", graph_id);
  else fprintf(file, "@g%d type xy\n", graph_id);

  // strcpy(x_label, xlabel);
  fprintf(file, "@ xaxis label \"%s\"\n", xlabel);
  fprintf(file, "@ xaxis label char size 1.00\n");
  fprintf(file, "@    legend off\n");

  for (int i = 0; i < set_num; i++)
    {
    fprintf(file, "@ s%d linestyle 1\n", i);
    fprintf(file, "@ s%d linewidth 2\n", i);
    fprintf(file, "@ s%d color %d\n", i, setColor(i));
    fprintf(file, "@ s%d symbol size 0.8\n", i);
    }
 
  if (set_num > 1)
    {
    double dx = 1.0 / double(cols);
    double dy = 1.0 / double(rows);  
    int row = graph_id / cols;
    int col = graph_id - row * cols;

    double ly = 1.0 - row * dy - dymargin / rows - dy_legend / rows;
    double lx = (col + 1) * dx - 0.5 * dxmargin / cols - dx_legend / cols;

    fprintf(file, "@    legend loctype view\n");
    fprintf(file, "@    legend layout 0\n");
    fprintf(file, "@    legend vgap 2\n");
    fprintf(file, "@    legend hgap 1\n");
    fprintf(file, "@    legend length 2\n");
    fprintf(file, "@    legend box on\n");
    fprintf(file, "@    legend box color 1\n");
    fprintf(file, "@    legend x1 %g\n", lx);
    fprintf(file, "@    legend y1 %g\n", ly);
    fprintf(file, "@    legend font 4\n");
    fprintf(file, "@    legend char size 1.000000\n");
    fprintf(file, "@    legend color 1\n");                                                      
    }
  }

//*******************************************************************************

int XmgrPlot::setColor(int set) {

  const  int setColors[30] = 
   {2, 4, 10, 1, 9, 11, 13, 3, 8, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7}; 

  return setColors[set];
}
//*******************************************************************************

void XmgrPlot::processStrings(TokenString &TS) {
  char   str[1024];
  double x=0.05, y=1.0, charSize = 0.8;
  int    sFont = 4, sColor= 4;

  for (int i=0; i< TS.token_count("plot.string"); i++) {
    y -= charSize * 0.03;
    TS.trail_pars("plot.string", i+1, 's', str, 'd', &x, 'd', &y, 'd', &charSize, 'i', &sFont, 'i', &sColor);
    fprintf(file, "@with string \n@string %g , %g \n@string char size %f \n",x, y, charSize);
    fprintf(file, "@string font %d \n@string color %d \n", sFont, sColor);
    fprintf(file, "@string on \n@string def \"%s\"\n", str);
  }
}


//*******************************************************************************
//               C L A S S   X M G R   P O I N T   P L O T
//*******************************************************************************

 XmgrPointPlot::XmgrPointPlot(int sets, char is_log, double T, const char *xlabel) : 
 XmgrPlot(sets, is_log, xlabel), PlotObj(0, 0, "", xlabel)
  {
  last = T;
  steps = XMGR_STEPS / cols;
  tScale = double( steps ) / last;

  tptrs = new double *[sets];
  fptrs = new double *[sets];
  x_old = new double[sets];
  f_old = new double[sets];
  x_temp = new double[sets];
  f_temp = new double[sets];

  pointCount = new int[sets];
  maxPoints  = new int[sets];
  tArray  = new double *[sets];
  fArray  = new double *[sets];

  fmin = 1e20; fmax = -1e20;

  tHold = fHold = 0.0;
  }
  
//*******************************************************************************

void XmgrPointPlot::get_range()
  {
  for (int i = 0; i < set_num; i++)
      {
      if ( *fptrs[i] < fmin )  fmin = *fptrs[i];
      if ( *fptrs[i] > fmax )  fmax = *fptrs[i];   
      }

  if ( (fabs((fmax - fmin)/(fmax + fmin + 1.0e-12)) < 1.0e-16) || (fmax <= fmin) ) 
      fmax = fmin + 1e-8;
  }
 
//*******************************************************************************

 void XmgrPointPlot::set_set(double *ptr, double *tptr, const char *txt)
    {
    fptrs[set_count] = ptr;
    tptrs[set_count] = tptr;
    x_old[set_count] = x_temp[set_count] = -100.0;
    f_old[set_count] = f_temp[set_count] = 0.0;

    pointCount[set_count] = 0;
    maxPoints[set_count]  = XMGR_STEPS;

    tArray[set_count] = new double[maxPoints[set_count]];
    fArray[set_count] = new double[maxPoints[set_count]];
    
    strcat(win_title, txt);

    fprintf(file, "@with g%d\n", graph_id);
    fprintf(file, "@g%d on\n", graph_id);
    fprintf(file, "@ legend string %d \"%s\"\n", set_count, txt);

    set_count ++;

    if (set_count < set_num) strcat(win_title, ", ");
    else
      {
      first = *tptr;
      last += first;
      char format[30]; 
      char str[50];
      double gx = format_double(first, last, XMGR_DIVISIONS, 2, format);

      sprintf(str, format, first);  fprintf(file, "@ world xmin %s\n", str);
      sprintf(str, format, last);   fprintf(file, "@ world xmax %s\n", str);
      fprintf(file, "@ xaxis tick major %g\n", gx);
      fprintf(file, "@ xaxis tick minor %g\n", gx/2.0);

      fprintf(file, "@ title \"%s vs. %s\"\n", win_title, x_label);
      fprintf(file, "@ title font 7\n");
      fprintf(file, "@ title size %g\n", 1.4 + 0.2 * (1 - cols) );
      fprintf(file, "@ subtitle font 4\n");
      fprintf(file, "@ subtitle size %g\n", 0.8 + 0.1 * (1 - cols) );
      fprintf(file, "@ title color %d\n", title_color);
      fprintf(file, "@ yaxis label \"%s\"\n", win_title);
      fprintf(file, "@ yaxis label char size 1.00\n");
      } 
    }

//*************************************************************************************
//                         X M G R   P L O T   D R A W
//*************************************************************************************


void  XmgrPointPlot::draw()
   {
   bool   redrawPlot = false, redrawFlag, newValue, drawTemp;
   bool   firstSet = true;
   double tiny = 1.0e-16;
   char   xtxt[40], mmform[40], mintxt[40], maxtxt[40], ytxt[40];
   char   format[80];

   for (int set = 0; set < set_num; set++)   {

     double Time  = *tptrs[set];
     double Val   = *fptrs[set];
     drawTemp = false;

     if (isnan(Val)) globalError(StrCpy("Undefined (nan) value passed to Xmgr point plot"));

     if ( Time < x_old[set] ) { 
       if ( firstSet )  { fmin = 1e20; fmax = -1e20; }
       truncate(set);
     }
     else {
       redrawFlag = ( int( Time * tScale ) == int( x_old[set] * tScale) ) ? false : true;
       newValue = ( fabs(f_old[set] - Val) < (fmax - fmin) * UPDATE_ACCURACY ) ? false : true;
       if ( !redrawFlag && !newValue ) { x_temp[set] = Time; f_temp[set] = Val; continue; }
       if (redrawFlag) redrawPlot = true;
       if (newValue && pointCount[set])
	 if (x_temp[set] > x_old[set])    // if we didn't draw this point before
           if ( fabs(f_temp[set] - Val) > (fmax - fmin) * UPDATE_ACCURACY ) { // and we deviated from it
             tArray[set][pointCount[set]] = x_temp[set];        // store it
             fArray[set][pointCount[set]] = f_temp[set];
             if (++pointCount[set] >= maxPoints[set]) addMemory(set);   
             drawTemp = true;
	     } 
     }

     double xDelta = fabs( Time - x_old[set] );
     //double yDelta = fabs( Val  - f_old[set] ); 

     tArray[set][pointCount[set]] = x_old[set] = Time;
     fArray[set][pointCount[set]] = f_old[set] = Val;
     if (++pointCount[set] >= maxPoints[set]) addMemory(set);   

     if ( firstSet ) 
       {
       get_range();
       
       double dy = format_double(fmin, fmax, XMGR_DIVISIONS, 2, format);
       double y0 = logScale ? fmin : dy * floor(fmin / dy); 
       double y1 = logScale ? fmax : dy * ceil(fmax / dy); 
       if ( logScale && !(y0 > 0) ) y0 = 1e-16;         
       if ( fabs( y1 - y0 ) < (tiny + dy) ) y1 = y0 + tiny + dy;
       
       format_double(Time-xDelta, Time, 1, 2, format);
       format_double(fmin, fmax, 1, 4, mmform);

       if (fmin == 0.0) strcpy(mintxt,"0"); else sprintf(mintxt, mmform, fmin);
       if (fmax == 0.0) strcpy(maxtxt,"0"); else sprintf(maxtxt, mmform, fmax);
       sprintf(xtxt, format,  Time);
       sprintf(ytxt,"%.5g",Val);

       fprintf(file, "@ WITH G%d\n", graph_id);
       fprintf(file, "@ G%d ON\n", graph_id);       
       fprintf(file, "@ subtitle \"%s at %s (min=%s max=%s)\"\n", ytxt, xtxt, mintxt, maxtxt);

       dy = format_double(y0, y1, XMGR_DIVISIONS, 2, mmform);
       sprintf(mintxt, mmform, y0);
       sprintf(maxtxt, mmform, y1);
       fprintf(file, "@ world ymin %s\n", mintxt);
       fprintf(file, "@ world ymax %s\n", maxtxt);
       fprintf(file, "@ yaxis tick major %g\n", dy);
       fprintf(file, "@ yaxis tick minor %g\n", dy*0.5);
       
       }
     firstSet = false;

     if (drawTemp) fprintf(file, "@ g%d.s%d POINT %.10g, %.10g \n", graph_id, set, x_temp[set], f_temp[set]);
     fprintf(file, "@ g%d.s%d POINT %.10g, %.10g \n", graph_id, set, Time, Val);
     x_temp[set] = Time; f_temp[set] = Val;
 
     /*{ sprintf(xtxt, format,  Time);
         format_double(Val-yDelta, Val, 1, 5, format);
         sprintf(ytxt, format, Val);
         fprintf(file, "@ g%d.s%d POINT %s, %s \n", graph_id, set, xtxt, ytxt ); }*/

     if ( (set + 1 == set_num) && (graph_id + 1 == graph_num) && redrawPlot && pointCount[set] > 1 )
       {
       fprintf(file, "@ redraw\n");
       fflush(file);
       }
      
     } // ** end loop over data sets
   }

//*************************************************************************************

void  XmgrPointPlot::truncate(int set)
   {
   fprintf(file, "@WITH G%d\n", graph_id);
   fprintf(file, "@KILL s%d\n",  set);
   fprintf(file, "@s%d color %d\n",  set, setColor(set));
   fprintf(file, "@s%d linewidth 2\n", set);

   int i = 0;
   double t, f;

   //char format[40], tStr[50], fStr[50];

   while ( ( tArray[set][i] < *tptrs[set] ) && ( i < pointCount[set] ) ) {
     //double dt = tArray[set][i] - t;
     //double df = fArray[set][i] - f;
     t = tArray[set][i]; f = fArray[set][i];
     if ( !isnan( f ) )  fprintf(file, "@ s%d POINT %.10g, %.10g \n", set, t, f );
     //if ( !isnan( f ) )  fprintf(stdlog, "@ s%d POINT %.10g, %.10g \n", set, t, f );
     /*  if (i == 0) fprintf(file, "@ s%d POINT %.8g, %.8g \n", set, t, f );
       else {
         format_double(t-dt, t, 1, 3, format); sprintf(tStr, format, t);
         format_double(f-df, f, 1, 3, format); sprintf(fStr, format, f);
         fprintf(file, "@ s%d POINT %s, %s \n", set, tStr, fStr );
       
      }
     */
     if ( f > fmax ) fmax = f;
     if ( f < fmin ) fmin = f;
     i++;
   }

   pointCount[set] = i;
   }

//*************************************************************************************

void  XmgrPointPlot::addMemory(int set)
   {
    int maxOld = maxPoints[set];
    int maxNew = maxPoints[set] += 400;
    double *tArrayNew = new double[maxNew];
    double *fArrayNew = new double[maxNew];

    for (int i = 0; i < maxOld; i++) 
      { tArrayNew[i] = tArray[set][i]; fArrayNew[i] = fArray[set][i]; }
 
    delete [] tArray[set];   delete [] fArray[set];
    tArray[set] = tArrayNew; fArray[set] = fArrayNew; 
   }


//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//*******************************************************************************
//               C L A S S   1 D   X M G R   P I P E   P L O T
//*******************************************************************************

Xmgr1Dplot::Xmgr1Dplot(FieldObj *f, char islog, const char *dir, double coord1, double coord2, 
                       double T, int sets) : XmgrPlot(sets, islog),
            FieldPlot1D(f, 0, dir, coord1, coord2)
{

  x_value = -100.0;
  last = T;
  steps = UPDATE_STEPS_1D;

  for (int set = 1; set < set_num; set++)  setTime[set] = x_value;

  fprintf(file, "@ title \"%s\"\n", win_title);
  fprintf(file, "@ title font 7\n");
  fprintf(file, "@ title size %g\n", 1.4 + 0.2 * (1 - cols) );
  fprintf(file, "@ subtitle font 4\n");
  fprintf(file, "@ subtitle size %g\n", 1.2 + 0.2 * (1 - cols) );
  fprintf(file, "@ title color %d\n", title_color);
  fprintf(file, "@ yaxis label \"[%s] (\\8m\\4M)\" \n", field->ID);
  fprintf(file, "@ yaxis label char size 1.00\n");
  fprintf(file, "@ xaxis label \"distance (\\8m\\4m)\" \n");
  fprintf(file, "@ xaxis label char size 1.00\n");
}

//*******************************************************************************

int Xmgr1Dplot::setColor(int set) {

 const  int setColors[30] = 
   {2, 2, 11, 5, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7}; 

 return setColors[set];
}

//*************************************************************************************
//                     1 D   X M G R   P L O T   D R A W
//*************************************************************************************

void  Xmgr1Dplot::draw()
   {
   bool   empty = true;
   double tScale = double(steps) / last;
   int    set;
   
   //fprintf(stdlog,"%s[%g -> %g] ", field->ID, x_value, field->Time);

   if ( field->Time < x_value )                   // pop the set stack
     while (setTime[set_num-1] > field->Time && setTime[set_num-2] < setTime[set_num-1])
       for (set = set_num-1; set > 0; set--) {
         fprintf(file, "@ move g%d.s%d to g%d.s%d \n", graph_id, set-1, graph_id, set);
         setTime[set] = setTime[set-1];
       }
   else if  ( int( field->Time * tScale + 1e-20) == int( x_value * tScale  + 1e-12) ) return;

   //  fprintf(stdlog,"\n");

   FieldPlot1D::get_range();

   fprintf(file, "@WITH G%d\n", graph_id);
   fprintf(file, "@G%d ON\n", graph_id);
   fprintf(file, "@ subtitle \"time = %gms\"\n", field->Time);   
   
   for (set = 1; set < set_num; set++) {                    // push set stack
     fprintf(file, "@ move g%d.s%d to g%d.s%d \n", graph_id, set, graph_id, set-1 );
     
     setTime[set-1] = setTime[set];
   }
   setTime[set_num-1] = x_value = field->Time;

   for (set = 0; set < set_num; set++) {
     fprintf(file,"@ s%d color %d\n", set, Xmgr1Dplot::setColor(set_num-1-set) ); 
     fprintf(file,"@ s%d linewidth 1\n", set);
     fprintf(file,"@ legend string %d \"t=%gms\" \n", set, setTime[set]);
   }
   fprintf(file,  "@ s%d linewidth 2\n", set_num-1);
   
   double y;
   //char format[40], xStr[50], yStr[50];

   int iLeft = 0, iRight = num-1;
   while ( !(field->ptype[field_index+iLeft*incr] & REGION_MASK) ) iLeft++;
   while ( !(field->ptype[field_index+iRight*incr] & REGION_MASK) ) iRight--;

   for (int j = iLeft; j <= iRight; j++) 
       if ( !isnan( y = get_value(field_index + incr * j) ) ) {
         empty = false;
       /*
       double x = coord[j], dx, dy;

       if (j == 0)  { dx = coord[1] - x ;   dy = get_value(ind+incr) - y; }
       else         { dx = x - coord[j-1] ; dy = y - get_value(ind-incr); }

       format_double(dx + 1e-9, 1, 5, format); sprintf(xStr, format, coord[j]);
       format_double(dy + 1e-9, 1, 5, format); sprintf(yStr, format, get_value(ind));
       fprintf(file, "@ g%d.s%d POINT %s, %s \n", graph_id, set_num-1, xStr, yStr );
       */
         fprintf(file, "@ g%d.s%d POINT %.10g, %.10g \n", graph_id, set_num-1, coord[j], y);
       }   
       else globalError(StrCpy("Undefined (nan) value passed to Xmgr 1D plot"));

   if ( !empty ) {  fprintf(file, "@ s%d on\n", set_num-1);
                    fprintf(file, "@ autoscale\n"); 
   }

   if (graph_id + 1 == graph_num)
         {
         fprintf(file, "@ redraw\n");
         fflush(file);
         }

   }   

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//*************************************************************************************
//                       U T I L I T Y   R O U T I N E S
//*************************************************************************************

double format_double(double number0, double number1, int divisions, int sigdigits, char *fstr)
{
const  int maxzeros = 4;
double dx, delta;
int    digits = 0, digitsBase = 0, zeros;

   strcpy(fstr,"%.");
   char sigdig[20];
   strcpy(sigdig,"");

   dx = fabs(number1 - number0) / double(divisions);
   delta = format_digits( dx, digits );
   double base = fabs(number1) > fabs(number0) ? fabs(number1) : fabs(number0);
   format_digits(base, digitsBase);

   if (dx < 1) {
     zeros = base < 1 ? digitsBase - 1 : 0;
     if (zeros <= maxzeros) sprintf(sigdig, "%df", digits + sigdigits - 1);
     else sprintf(sigdig, "%de", digits + sigdigits - zeros - 2);
   } 
   else {
     zeros = digits - sigdigits;
     if (zeros <= maxzeros) {
        int afterDec =  zeros < 0 ? -zeros : 0 ;
        sprintf(sigdig, "%df", afterDec);
     }
     else sprintf(sigdig, "%de", digitsBase - zeros - 1); 
   }

  strcat(fstr, sigdig);
  //fprintf(stderr, "n0=%g n1=%g digitsBase=%d dx=%g digits=%d sigdigits=%d zeros=%d delta=%g format=%s\n",
  //                 number0, number1, digitsBase, dx, digits,sigdigits,zeros,delta,fstr);
  return delta;
}
//*************************************************************************************
// Returns the number of digits in "dx", before or after the decimal point

double format_digits(double dx, int &digits)
{
double mult = 1;
int    grid;
 double tiny = 1e-8;

   digits = 0;

   //fprintf(stderr, "dx=%g ",dx);
   if (dx < 1) 
     { 
     while ( dx * mult + tiny < 1 && dx * mult > 0.0 ) { mult *= 10; digits++; } 
     grid = int(dx * mult + 0.5);
     if ((grid & 1) && (grid != 1)  && (grid != 5)) grid--;  // prefer even grid
     dx = double(grid) / double(mult); 
     }
   else if (dx >= 1)
     { 
     while ( dx / mult >= 10) { mult *= 10; digits++; }
     grid = int(dx / mult + 0.5);
     if ((grid & 1) && (grid != 1)  && (grid != 5)) grid--;   // prefer even grid
     dx = double(grid) * double(mult); 
     digits++;
     }

   //fprintf(stderr, "digits=%d dx=%g grid=%d mult=%g\n",digits,dx,grid,mult);
   return dx;
}

//*************************************************************************************
//   T O K E N - S T R I N G   C O N S T R U C T O R   F O R   P L O T   A R R A Y
//*************************************************************************************
  
int PlotArray::get_plot_num(TokenString &TS)
{
  int graphs = 0;

  int Total  = TS.token_count("plot") + TS.token_count("Export");

  int mutes  = TS.token2_count("plot","mute") + TS.token2_count("plot","mute.log") +
               TS.token2_count("plot","1D.mute") + TS.token2_count("plot","1D.mute.log") +
               TS.token2_count("plot","2D.mute") + TS.token2_count("plot","2D.mute.log") +
               TS.token2_count("plot","dump")    + TS.token_count("Export");
  //  int npoint = TS.token2_count("plot","point")   + TS.token2_count("plot","point.log");
  int n1D    = TS.token2_count("plot","1D")      + TS.token2_count("plot","1D.log");
  int n2D    = TS.token2_count("plot","2D")      + TS.token2_count("plot","2D.log");
  int nraster= TS.token2_count("plot","raster")  + TS.token2_count("plot","raster.log");
  
  int npoint = Total - mutes - n1D - n2D - nraster;

  char smethod[10];
  strcpy(smethod,"");
  
  
  if (TS.token_count(LOOP_TOKEN)) { 
    graphs = 0; 
    method = METHOD_MUTE; 
    }
  else  {
      if (TS.token2_count("plot.method","xmgr")) {  
      int rows = 0, cols = 0;
      TS.trail_pars("plot.method", 1, 0,0, 'i', &rows, 'i', &cols);
      method = METHOD_XMGR;
      strcpy(smethod,"Xmgr");
      XmgrPlot::init(TS, npoint + n1D, stdout, rows, cols);
      graphs = npoint + n1D;
    }
    else if (TS.token2_count("plot.method","gl")) 
      {
      graphs = npoint + n1D + n2D + nraster;
      strcpy(smethod,"gl");
      method = METHOD_YGL;
      }
    else method = METHOD_MUTE;
  }

  if (TS.token_count("plot.print")) {
    long pos;
    mutes += n1D + n2D;
    if ( method != METHOD_XMGR ) mutes += npoint;
    else for(int i=1; i<=TS.token_count("plot"); i++) 
      if ( !equal_( TS[ pos = TS.token_index("plot", i) + 1 ], "1D") &&
           !equal_( TS[ pos = TS.token_index("plot", i) + 1 ], "2D") &&
           !equal_( TS[ pos = TS.token_index("plot", i) + 1 ], "raster") &&
           !equal_( TS[ pos = TS.token_index("plot", i) + 1 ], "dump") &&
           !equal_( TS[ pos = TS.token_index("plot", i) + 1 ], "mute") ) 
         mutes += TS.tokens_to_eol(pos);
  }


  int total = graphs + mutes;

  if (VERBOSE) 
   fprintf(stdlog,"\n### Setting up %d plots (%d %s graphs plus %d disc dumps):\n", 
           total, graphs, smethod, mutes);

  return  total;
}

//*****************************************************************************
//                      S T A T U S      W I D G E T
//*****************************************************************************

  RunStatusString::RunStatusString(int length, const char *s1, double *p1, double T, char *s2, double *p2) {
    widgetLength = length;
    if (widgetLength < 3) widgetLength = 1;
    widgetStr = new char[length+1];
    timeStr = StrCpy(s1);   
    timePtr = p1; 
    realTime = double( clock() / CLOCKS_PER_SEC );
    if (s2) 
      { extraStr = StrCpy(s2);  extraPtr = p2;  wholeStr = new char[length+strlen(s1)+strlen(s2)+60];}
    else { wholeStr = new char[length+strlen(s1)+60]; extraStr = 0; extraPtr = 0; }
    reset(T);
    //update();
  };

//*****************************************************************************

  RunStatusString::~RunStatusString() {
    fprintf(stdlog,"\n");
    delete [] widgetStr; delete [] wholeStr; 
    delete [] timeStr; if (extraStr) delete [] extraStr;
  };

  
//*****************************************************************************

  void RunStatusString::reset(double T) {
    widgetCount = 0;
    strcpy(widgetStr,"");
    for (int i = 0; i < widgetLength; i++) strcat(widgetStr,"_");
    if (timePtr) {
      t0 = *timePtr;
      if (T > 0) total = T - t0;
    }
    strcpy(wholeStr, "");
  }    

//*****************************************************************************

  void RunStatusString::update(char c) {
    int i;

    double newTime = double( clock() / CLOCKS_PER_SEC );
    if (newTime < realTime + 1) return;
    realTime = newTime;

    for (i = strlen(wholeStr); i > 0; i--) fprintf(stdlog,"\b");

    if (widgetLength < 3) {
        char clockchar[4] = { '-', '\\', '|', '/' };
        widgetCount = ++widgetCount % 4;
        widgetStr[0] = clockchar[widgetCount];
     }
     else if ( c != 0) {
        if (widgetCount < widgetLength) widgetStr[widgetCount++] = c;
        else {
          for (i = 1; i < widgetLength ; i++ ) widgetStr[i-1] = widgetStr[i];
          widgetStr[widgetLength - 1] = c;
        }
     }

    //fprintf(stdlog,"hello?");
    if (extraStr) 
      sprintf(wholeStr," [%s] (%d %%) %s=%.6g, %s=%.3g       ", widgetStr, int(100*(*timePtr-t0)/total + 0.5), 
	 	          timeStr, *timePtr, extraStr, *extraPtr);
    else if (timePtr) 
       sprintf(wholeStr," [%s] (%d %%) %s=%.6g       ", widgetStr, int(100*(*timePtr-t0)/total + 0.5),
                          timeStr, *timePtr ); 
    else
       sprintf(wholeStr,"%s", widgetStr);    

    fprintf(stdlog,"%s",wholeStr); 
  }


//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//*************************************************************************************

PlotArray::PlotArray(SimulationObj &SO)
{
  int      hor = pref_width, ver = pref_height;
  long     POS;
  char     prefix[128], fileName[255];  
  double   SimTime = SO.totalSimTime;
  int      vcount = 0;  // counter for the verbose comment print


  count = 0; 

  TokenString *params = SO.Params;
  plot_num = get_plot_num( *params );

  if (!plot_num) return;
  array   = new PlotObj *[plot_num];

  params->get_int_param("plot.gl.width", &hor);
  params->get_int_param("plot.gl.height", &ver);
  params->get_int_param("plot.point.steps", &PlotObj::UPDATE_STEPS);
  params->get_int_param("plot.1D.steps", &PlotObj::UPDATE_STEPS_1D);
  params->get_param("plot.update.accuracy", &PlotObj::UPDATE_ACCURACY);
  ygl_on = 0;

  strcpy(fileName,"");
  strcpy(prefix,"temp");
  if ( params->token_count("plot.print", &POS) )  params->line_string(POS + 1, prefix);

  //*************** Separate block for state export plot ****************************

  for (int ii = 1; ii <= params->token_count("Export"); ii++) {
      double tDump = SimTime;
      params->trail_pars("Export", ii, 'd', &tDump, 's', fileName);
      array[count++] = new StateDump(&SO, tDump, fileName);
      if (VERBOSE)  
        fprintf(stdlog," plot #%d: State variable dump at t=%g into file \"%s\"\n", count, tDump, fileName);
      vcount++;
    }

  //*********************************************************************************

  for (int i = 1; i <= params->token_count("plot"); i++) 
  try
    {
    char     ptype[100], VarID[200], log_plot;

    POS = params->token_index("plot", i) + 1; 

    params->get_string(POS, ptype); 
    int pp = strlen( ptype );
    if ( equal( ptype+pp-4, ".log" ) ) { log_plot = 1; ptype[pp-4] = '\0'; }
    else if ( equal( ptype, "log" ) ) { log_plot = 1; strcpy(ptype,"point"); }
    else log_plot = 0;

   if ( equal(ptype, "point") || equal(ptype, "mute") || SO.ResolveID(ptype) ) {  

      double   *time_ptr=0, *kin_ptr=0;
      int set_num = 1, xmgr_plot;

      if ( SO.ResolveID(ptype) ) POS--;

      if ( ( method == METHOD_XMGR ) && !equal(ptype, "mute") ) {
        set_num = params->tokens_to_eol(POS+1);
        xmgr_plot = count;
        array[count++] = new XmgrPointPlot(set_num, log_plot, SimTime);
      } 

     //******************  MULTIPLE SETS IF XMGR POINT PLOT

    for (int set = 0; set < set_num; set++)
      { 
      strcpy(VarID, "(null)");
      POS += params->trail_pars(POS, 'S', VarID, 'E');
      if ( ! ( kin_ptr = SO.ResolveID( VarID, &time_ptr) ) )
        params->errorMessage(POS, makeMessage("Undefined variable \"%s\" in plot statement", VarID) );

      if ( ( method == METHOD_XMGR ) && !equal(ptype, "mute") ) 
        ((XmgrPointPlot *)array[xmgr_plot])->set_set(kin_ptr, time_ptr, VarID);
      else if ( (method == METHOD_YGL) && !equal(ptype, "mute") ) // set_num = 1
        {
        ygl_on = 1;
        params->trail_pars(POS, 'i', &hor, 'i', &ver,  'E');
        array[count++] = new PointPlot(kin_ptr, time_ptr, log_plot, SimTime, VarID, "time=", hor, ver);
        }
      if ( params->token_count("plot.print") || equal(ptype,"mute") )
        {
        if ( equal(ptype,"mute") ) params->line_string(POS + 1, fileName);
        else sprintf(fileName,"%s%s", prefix, VarID);
        array[count++] = new MutePointPlot(kin_ptr, time_ptr, log_plot, SimTime, fileName, VarID);
	}
      }  //@@@@@@@@@@@@@@@@@  END SET CYCLE OVER POINT PLOT SETS @@@@@@@@@@@@@@@@@@@

    }  else {  // not a point plot

    FieldObj *field = 0;
    double   coord1=0, coord2=0;
    char     dir[MAX_TOKEN_LENGTH];
    strcpy(fileName,"");     

    strcpy(VarID, "(null)");
    POS += params->trail_pars(POS, 'S', VarID, 'E');
    if ( !( field = SO.ResolveField(VarID) ) ) 
       params->errorMessage(POS, makeMessage("Undefined field \"%s\" in plot statement",VarID) );

    if ( equal_(ptype,"2D") ) {
      if (DIMENSIONALITY < 2) throw StrCpy("Cannot make a 2D plot of a 1-dimensional field");
      strcpy(dir, "z"); // default direction for 2D plots in 2D geometry 
      coord1 = 0.0;
      if (DIMENSIONALITY > 2) POS += params->trail_pars(POS, 'S', dir, 'd', &coord1, 'E');
      if ( equal(ptype,"2D")  && (method == METHOD_YGL) ) {
        ygl_on = 1;
        params->trail_pars(POS, 'i', &hor, 'i', &ver);
        array[count++] = new Plot2D(field, log_plot, dir, coord1, SimTime, hor, ver);
        }
      if ( equal(ptype, "2D.mute") || params->token_count("plot.print") ) {
        double tt = SimTime;
        if ( equal(ptype,"2D.mute") )  params->trail_pars(POS, 'd', &tt, 's', fileName, 'E');
        array[count] = new MutePlot2D(field, log_plot, dir, coord1, tt, fileName, VarID);
        if ( !equal(ptype,"2D.mute") ) { sprintf(fileName, "%s%s", prefix, array[count]->win_title);
                                         strcpy( ((MutePlot2D *)array[count])->fileName, fileName); }
        count++;
      }
    }
    else if ( equal(ptype,"dump") ) {
      double tt = SimTime;
      params->trail_pars(POS, 'd', &tt, 's', fileName);
      array[count++] = new FieldDump(field, tt, fileName);
    }
    else if ( equal_(ptype,"1D") || ( equal(ptype,"raster") && (method == METHOD_YGL) ) ) {
      strcpy(dir, LABEL_DIM1);  // default direction for 1D plots in 1D geometry
      coord1 = coord2 = 0;
      if (DIMENSIONALITY == 3) POS += params->trail_pars(POS, 'S', dir, 'd', &coord1, 'd', &coord2, 'E');
      else if (DIMENSIONALITY == 2) POS += params->trail_pars(POS, 'S', dir, 'd', &coord1, 'E');
      if ( equal(ptype,"raster") && (method == METHOD_YGL) ) {
         ygl_on = 1;
         params->trail_pars(POS, 'i', &hor, 'i', &ver, 'E');
         array[count++] = new RasterPlot(field, log_plot, dir, coord1, coord2, SimTime, hor, ver);
         }
      else if ((method == METHOD_YGL) && equal(ptype,"1D")) {
        ygl_on = 1;
        params->trail_pars(POS, 'i', &hor, 'i', &ver,  'E');
        array[count++] = new Plot1D(field, log_plot, dir, coord1, coord2, SimTime, hor, ver);
        }
      else if ((method == METHOD_XMGR) && equal(ptype,"1D")) {
        int sets = 5;
        params->trail_pars(POS, 'i', &sets, 'E');
        array[count++] = new Xmgr1Dplot(field, log_plot, dir, coord1, coord2, SimTime, sets);
      } 
      if ( params->token_count("plot.print") || equal(ptype,"1D.mute")) 
        {
        if ( equal(ptype,"1D.mute") )  params->get_string( POS+1, fileName );
        array[count] = new MutePlot1D(field, log_plot, dir, coord1, coord2, SimTime, fileName, VarID);
        if ( !equal(ptype,"1D.mute") ) { sprintf(fileName, "%s%s", prefix, array[count]->win_title);
	                                 strcpy( ((MutePlot1D *)array[count])->fileName, fileName); }
        count++;
	}
      } // endif 1D plot
    else {         // ignore some plots depending on plot_method
      fprintf(stdlog," *** ignoring \"%s\" plot\n", ptype);
      continue;
      }
    } // end if not point plot
  
    if (VERBOSE) for(; vcount < count; vcount++) {
      char s[5];
      if (array[vcount]->log_plot) strcpy(s,"log "); else strcpy(s,"");
      fprintf(stdlog," plot #%d: %splot of %s\n", vcount+1, s, array[vcount]->win_title);
    }
  } catch (char *str) { 
      params->errorMessage( params->token_index("plot", i) + 1, str, "Invalid plot statement"); 
    }
    catch (int m) { 
      params->errorMessage( params->token_index("plot", i) + 1, 0, "Invalid plot statement"); 
    }
      // ****************   end loop over plot number
}

//*************************************************************************************
//*************************************************************************************
  
PlotArray::~PlotArray()  { 

  if (!count) return;

  tail_draw(); 

  for (int i = 0; i < count; i++) delete array[i]; 

  delete [] array;

}

//*************************************************************************************
//*************************************************************************************
